// SPHGEN_CPP main.cpp
//
// Andrew Magis <amagis@cise.ufl.edu>
// Last change: Time-stamp: <26-Jul-2006 10:26:35 amagis>
//
// Version 1.1
// This software is open source, to be used in conjunction with the UCSF DOCK
// package.  If you have questions, comments, or suggestions, please contact
// me at the email address above.
//

#include <stdlib.h>
#include <unistd.h>
#include "Sphere.hpp"

//Global variables
Atom **atoms;
int num_atoms = 0;
int num_spheres = 0;
int num_clusters = 0;

int reader(char surftp, char dentag, char *msfilename);
int sphere(double dotlim, double radmin, double radmax, char surftp);
int cluster(FILE *outfile);
void usage();

int main(int argc, char **argv)
{
	FILE *infile = NULL, *outfile = NULL;

	//Set default inputs
	char msfilename[200] = "", outputfilename[200] = "";
	char surftp = 'R', dentag = 'X';
	double dotlim = 0.0, radmax = 4.0, radmin = 1.4;
	int items_read, command_line = 0, ch;

	//Try to read command-line options
	while ((ch = getopt(argc, argv, "i:s:d:l:x:m:o:")) != -1) {

		switch(ch) {
			
		case 'i':
			strcpy(msfilename, optarg);
			break;
		case 's':
			surftp = *optarg;
			if ((surftp != 'R') && (surftp != 'L')) {
				printf("Invalid surface topology option: %c\n", surftp);
				usage();
			}
			break;
		case 'd':
			dentag = *optarg;
			if (dentag != 'X') {
				printf("Invalid surface point option: %c\n", dentag);
				usage();
			}
			break;
		case 'l':
			dotlim = strtod(optarg, NULL);
			break;
		case 'm':
			radmin = strtod(optarg, NULL);
			break;
		case 'x':
			radmax = strtod(optarg, NULL);
			break;
		case 'o':
			strcpy(outputfilename, optarg);
			break;
		}
		command_line = 1;
	}

	if (!command_line) {

		printf("Reading from file INSPH\n");
		
		//Open infile INSPH
		infile = fopen("INSPH", "rt");
		if (infile == NULL) {
			fprintf(stderr, "ERROR! Cannot open file: INSPH");
			usage();
			exit(1);
		}
	
		items_read = fscanf(infile, "%s\n%c\n%c\n%lf\n%lf\n%lf\n%s",
							msfilename, &surftp, &dentag, &dotlim,
							&radmax, &radmin, outputfilename);

		//Close infile
		fclose(infile);
	}

	if (strlen(msfilename) == 0) {
		printf("Error! Input filename '%s' is invalid.\n", msfilename);
		usage();
	}
	else if (strlen(outputfilename) == 0) {
		printf("Error! Output filename '%s' is invalid.\n", outputfilename);
		usage();
	}
	
	//Now write this information to the screen:
	printf("Input file: %s\n", msfilename);
	printf("Surface Topology: %c\n", surftp);
	printf("Surface Point Subset: %c\n", dentag);
	printf("Min Distance: %.1f\n", dotlim);
	printf("Min Radius: %.3f\n", radmin);
	printf("Max Radius: %.3f\n", radmax);
	printf("Output file: %s\n", outputfilename);

	//Call function to read in the molecular surface file
	reader(surftp, dentag, msfilename);

	//Call sphere function to generate spheres
	sphere(dotlim, radmin, radmax, surftp);

	//Open outfile
	outfile = fopen(outputfilename, "wt");
	
	//Create clusters from the generated spheres
	cluster(outfile);

	//Close outfile
	fclose(outfile);

	//At the end of the program, delete the array of atoms
	for (int i = 0; i < num_atoms; i++)
		if (atoms[i] != NULL)
			delete atoms[i];

	if (atoms != NULL)
		delete[] atoms;
	return 0;
}

int reader(char surftp, char dentag, char *msfilename)
{

	FILE *infile = NULL;
	int num_surfacepoints, total_surfacepoints;
	int first, second, items_read;
	char residue_name[3], atom_name[3], tag[3], residue[4], temp[4];
	double not_used, mul = 1;
	
	//Position coordinates and normal vector 
	double x, y, z, nx, ny, nz, atomx, atomy, atomz;

	//Set the multiplier to be -1 if receptor is to be ligand
	if (surftp == 'L')
		mul = -1;
	
	//Open the molecular surface file
	infile = fopen(msfilename, "rt");
	if (infile == NULL) {
		fprintf(stderr, "ERROR! Cannot open file: %s\n", msfilename);
		exit(1);
	}
	
	//We are trying to avoid setting an upper limit on lines
	//in the program, so before allocation, count number of atoms
	num_atoms = 0;
	while (!feof(infile)) {
		items_read = fscanf(infile, "%s%s%s%lf%lf%lf%s%lf%lf%lf%lf\n",
							residue_name, residue, atom_name, &x, &y,
							&z, tag, &not_used, &nx, &ny, &nz);
		if (tag[0] == 'A')
			num_atoms++;
	}

	rewind(infile);

	//Allocate array of atoms
	atoms = new Atom*[num_atoms];

	total_surfacepoints = 0;
	//Now parse through the file and extract the surface points
	for (int i = 0; i < num_atoms; i++) {
		
		//Get initial atom information
		items_read = fscanf(infile, "%s%s%s%lf%lf%lf%s%lf%lf%lf%lf\n",
							residue_name, residue, atom_name, &x, &y,
							&z, tag, &not_used, &nx, &ny, &nz);

		//Keep running sum of all surface points
		num_surfacepoints = 0;
		first = ftell(infile);
		do {

			//Count surface points that follow this atom
			items_read = fscanf(infile, "%s%s%s%lf%lf%lf%s%lf%lf%lf%lf\n",
								temp, temp, temp, &atomx, &atomy,
								&atomz, tag, &not_used, &nx, &ny, &nz);
			num_surfacepoints++;

		} while ((tag[0] != 'A') && (!feof(infile)));

		//If actually the end of file, keep the last line read
		if (feof(infile)) num_surfacepoints++;
		total_surfacepoints += num_surfacepoints-1;
		
		//Allocate atom object and pass it the number of surface points
		atoms[i] = new Atom(atomx, atomy, atomz, residue_name, residue,
							atom_name, num_surfacepoints-1, i+1);

		//Seek back to actually read in the file information
		second = ftell(infile);
		fseek(infile, first-second, SEEK_CUR);
		
		//Finally, parse through the surfacepoints once more to add the data
		for (int j = 0; j < num_surfacepoints-1; j++) {

			//Really read in data to be written
			items_read = fscanf(infile, "%s%s%s%lf%lf%lf%s%lf%lf%lf%lf\n",
								residue_name, residue, atom_name, &x, &y,
								&z, tag, &not_used, &nx, &ny, &nz);
			
			//Add the surface point to the atom object
			if(!atoms[i]->AddSurfacePoint(tag, x, y, z, nx*mul, ny*mul, nz*mul)) {
				fprintf(stderr, "ERROR! Unable to add surface point\n");
				exit(1);
			}
		}		
	}
	printf("%d Atoms, %d Surface Points\n",
		   num_atoms, total_surfacepoints);

	//Close the molecular surface file
	fclose(infile);
	return 1;
}

int sphere(double dotlim, double radmin, double radmax, char surftp)
{

	num_spheres = 0;

	//Compare each atom to each other atom
	for (int i = 0; i < num_atoms; i++) {
		
		for (int j = 0; j < num_atoms; j++) {
			
			//Compare the surface points of each atom
			if (atoms[i] != atoms[j])
				atoms[i]->CompareAtoms(atoms[j]);	
		}

		//After each atom completes, check to see if any spheres
		//should be created for any of the surface points on that atom
		num_spheres += atoms[i]->GenerateSpheres(dotlim, radmin, radmax);
	}

	//Spheres finished generating
	printf("Finished generating spheres\n");
	return 1;
}

//Sorts the array of spheres into clusters
int cluster(FILE *outfile)
{

	int index = 0, total_spheres = 0;
	double distance;
	Sphere *head = NULL, *tail = NULL, *u, *p;
	Sphere **clusters = new Sphere*[num_spheres];
	
	//Allocate array of pointers to point to spheres
	Sphere **spheres = new Sphere*[num_spheres]; 

	//Retrieve all the spheres that will be included in clusters
	for (int i = 0; i < num_atoms; i++) {

		//If index == num_spheres then all of the largest spheres
		//have been retrieved.  
		if (index >= num_spheres) {
			break;
		}

		total_spheres += atoms[i]->GetNumSpheres();
		spheres[index] = atoms[i]->GetLargestSphere();
		if (spheres[index] != NULL) {
			
			//Allocate set of pointers in each sphere
			spheres[index]->pairs = new Sphere*[num_spheres];
			spheres[index]->num_pairs = 0;
			spheres[index]->cluster_size = 0;
			
			//"Color" each sphere white (true) in preparation for BFS
 			spheres[index]->is_white = true;
			spheres[index]->next = NULL;
			index++;
		}
	}

	//Create graph of spheres
	for (int i = 0; i < num_spheres; i++) {
		for (int j = i+1; j < num_spheres; j++) {

			//Calculate Cartesian distance between spheres
			distance = sqrt( (*spheres[i]->cs - *spheres[j]->cs)
							 * (*spheres[i]->cs - *spheres[j]->cs) );

			//If the two spheres are touching, make a pair
			if (distance <= (spheres[i]->rtmin + spheres[j]->rtmin)) {
				spheres[i]->pairs[spheres[i]->num_pairs++] = spheres[j];
				spheres[j]->pairs[spheres[j]->num_pairs++] = spheres[i];
			}
		}
	}

	for (int i = 0; i < num_spheres; i++) {
		
		//If this sphere remains undiscovered by the algorithm
		if (spheres[i]->is_white) {
		
			//Set the head of the queue to be the first sphere
			head = spheres[i];
			tail = spheres[i];
			tail->prev = NULL;
			head->is_white = false;

			//p points to the most recent item in this cluster
			clusters[num_clusters] = head;
			clusters[num_clusters]->next = NULL;
			
			//Use breadth-first search to find clusters
			while (head != NULL) {
				
				u = head;	
				for (int j = 0; j < u->num_pairs; j++) {
					//If the color of this sphere is "white"
					if (u->pairs[j]->is_white) {
						
						//Color it "grey" and add to queue
						u->pairs[j]->is_white = false;
						tail->prev = u->pairs[j];
						u->pairs[j]->next = tail; 
						tail = u->pairs[j];
						tail->prev = NULL;
					}
				}
				//Dequeue the element at the front of the queue
				head = head->prev;

				p = clusters[num_clusters];
				
				//if cluster size is 0
				if (clusters[num_clusters]->cluster_size == 0) {
					//This sphere goes at the beginning
					if (u->this_atom < p->this_atom) {
						u->next = p;
						u->prev = NULL;
						p->prev = u;
						clusters[num_clusters] = u;
					}
					//This sphere goes at the end
					else {
						p->next = u;
						u->prev = p;
						u->next = NULL;
					}
				}
				else {	
					//Find the location in this cluster to insert this sphere
					while ((p->this_atom < u->this_atom) &&
						   (p->next != NULL))
						p = p->next;
					
					//If we are at the end of the linked-list
					if (p->next == NULL) {

						//This sphere goes on the end
						if (p->this_atom < u->this_atom) {
							p->next = u;
							u->prev = p;
							u->next = NULL;
						}
						//This sphere goes just before the end
						else {
							p->prev->next = u;
							u->prev = p->prev;
							p->prev = u;
							u->next = p;
						}	
					}
					else {
						//Insert this sphere
						p->prev->next = u;
						u->prev = p->prev;
						p->prev = u;
						u->next = p;
					}
				}
				//Increment the size of this cluster
				clusters[num_clusters]->cluster_size++;
			}
			//Only create a cluster if it has at least two spheres
			if (clusters[num_clusters]->cluster_size > 1)
				num_clusters++;
		}
	} //End of FOR loop

	//Sort the clusters by increasing size
	for (int i = 0; i < num_clusters; i++) {
		for (int j = i+1; j < num_clusters; j++) {
			if (clusters[j]->cluster_size > clusters[i]->cluster_size) {
				p = clusters[i];
				clusters[i] = clusters[j];
				clusters[j] = p;
			}
		}
	}
	
	printf("Clustering Complete: %d Clusters\n", num_clusters);

	fprintf(outfile, "DOCK receptor spheres (C++ version)\n");
	for (int i = 0; i < num_clusters; i++) {

		p = clusters[i];
		fprintf(outfile, "cluster%6d   number of spheres in cluster%6d\n",
			i+1, p->cluster_size);
		
		while (p != NULL) {

			//Print out the sphere information in clusters
			fprintf(outfile, "%5d%10.5f%10.5f%10.5f%8.3f%5d%2d%3d\n",
					p->this_atom, p->cs->x, p->cs->y, p->cs->z, p->rtmin,
					p->pair_atom, 0, 0);
			p = p->next;
		}
	}

	//Now print out all the spheres generated, including those not
	//placed in a cluster (not the largest per atom)
	fprintf(outfile, "cluster%6d   number of spheres in cluster%6d\n",
			0, total_spheres);
	for (int i = 0; i < num_atoms; i++) {
		for (int j = 0; j < atoms[i]->GetNumSpheres(); j++) {
			p = atoms[i]->GetSphere(j);
			fprintf(outfile, "%5d%10.5f%10.5f%10.5f%8.3f%5d%2d%3d\n",
					p->this_atom, p->cs->x, p->cs->y, p->cs->z, p->rtmin,
					p->pair_atom, 0, 0);
		}
	}

	//Individual spheres are deleted by the Atom class
	delete[] spheres;
	delete[] clusters;
	return 1;
}

//Prints usage information
void usage()
{
	printf("\nUsage: sphgen_cpp -i inputfilename [-s surface_topology] [-d surfacepoints] [-l min_distance] [-m min_radius] [-x max_radius] -o outputfilename\n\n");
	printf("\t-i\tInput file name [required]\n");
	printf("\t-o\tOutput file name [required]\n");
	printf("\t-s\tR: outside of receptor  L: inside of receptor [default: R]\n");
	printf("\t-d\tX: all surface points [default: X]\n");
	printf("\t-l\tMinimum distance between spheres [default: 0.0]\n");
	printf("\t-m\tMinimum sphere radius [default: 1.4]\n");
	printf("\t-x\tMaximum sphere radius [default: 4.0]\n\n");
	printf("If no command line options are specified, sphgen_cpp reads parameters from file INSPH in the working directory.\n");
	exit(1);
}



